%% tclp__type_order.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__type_order.pl,v 1.11 2004/01/26 19:40:23 manu Exp $

%% Handling of order between type constructors

%% Verbose modes debug_type_order

:- module(tclp__type_order, [ add_type/1, 
                              add_type/2,
                              add_order/2,
                              build_order/0,
                              tclp__glb/3,
                              tclp__lub/3,
                              tclp__break/2,
                              tclp__ordered_types/2,
                              type_exists/1,
                              init_order/0,
                              tclp__enumerate_tops/1,
                              tclp__enumerate_bottoms/1,
                              tclp__elim_arg_down/3,
                              tclp__elim_arg_up/3,
                              add_type_variable_constant/1,
			      is_type_variable_constant/1
                            ]).

:- use_module(library(chr)).
:- use_module(library(ordsets)).
:- use_module(library(lists)).
:- use_module(tclp(tclp__utils)).
:- use_module(tclp(tclp__arity)).
:- use_module(tclp(tclp__def)).
:- use_module(tclp(tclp__errors)).
:- use_module(tclp(tclp__arrays)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Order handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% We use a CHR handler to deal with the order
handler order.

%% Constraints over the order

constraints lt/2, type/1. %, invalidate/0, revalidate/0.

%% Rules

%% On type with a given name
elim @ type(X) \ type(X) <=> true.

%% When a new type is added, we say that it is smaller that top
%% and greater than top
add_implicit @ type(X) ==> 
        builtin_type(top, Top), lt(X,Top),
        builtin_type(bot, Bot), lt(Bot,X),
        add_term_ineq(X).


add_term_ineq(_Type) :- option(disableterm),!.
add_term_ineq(Type) :- builtin_type(top, Top),Top==Type,!.
add_term_ineq(Type) :- builtin_type(term,Term), lt(Type,Term).

%% We also add tau < term except for top and for type variables

%% 
idn @ lt(X,X) <=> true.

%% Idempotence
idem @ lt(X,Y) \ lt(X,Y) <=> true.

%% Transitivity
transitivity @ lt(X,Y), lt(Y,Z) ==> lt(X,Z).

%% We add top and bottom
init_order :-
        (option(disableterm) ->
            verbose(debug_type_order,['option disableterm enabled']);
            builtin_type(term,Term), add_type(Term)),
        builtin_type(top, Top), add_type(Top),
        builtin_type(bot, Bot), add_type(Bot),
        verbose(debug_type_order,'add builtin top & bottom ok'),
        add_all_types_chr,
        verbose(debug_type_order,'add types ok'),
%        (option(complete_order) -> complete_order ; true),
	complete_order,
        verbose(debug_type_order,'complete_order ok'),
        add_all_types_chr,
        verbose(debug_type_order,'add new introduced types ok'),
        add_all_order_decs,
        verbose(debug_type_order,'add order ok').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Computing greatest lower/lowest higher bounds method 2 : using an
%% appropriate table and data structure
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% first we need to associate to each type a number (its Id)
:- dynamic type_id/2, id_type/2.

%% find all declared types and associate to each one an id by asserting a
%% type_id clause and an id_type clause. Nb is the number of types.
associate_id_to_types(Nb) :-
        chr:findall_constraints(type(_),CstrList),
        retractall(type_id(_,_)),
        retractall(id_type(_,_)),
        length(CstrList, Nb),
        associate_all(CstrList,1).

%% associate to each contraint of the given list an id by asserting a
%% type_id clause and an id_type clause.
associate_all([],_).
associate_all([type(Type) # _| Constraints], Id) :-
        assert(type_id(Type,Id)),
        assert(id_type(Id,Type)),
        Id2 is Id + 1,
        associate_all(Constraints, Id2).
        
%% makes an order matrix for quick lookup. Matrix[i,j] == 1 iff the i <type j
make_order_matrix(NbTypes,Matrix) :-
        verbose(debug_type_order,'make_order_matrix'),
        create_matrix(NbTypes,Matrix),
        verbose(debug_type_order,'create matrix ok'),
        chr:findall_constraints(lt(_,_),Constraints),
        verbose(debug_type_order,'find constraints ok'),
        add_constraints_in_matrix(Constraints, Matrix),
        verbose(debug_type_order,'add constraints ok'),
        array_iteri_goal(Matrix,I,Column,
                         (array_element_at(Column,I,1)),
                         []),
        verbose(debug_type_order,'make_order_matrix done').

add_constraints_in_matrix([],_).
add_constraints_in_matrix([lt(T1,T2)#_|Rest], Matrix) :-
        (type_id(T1,Id1);print_error(['Unknown type ',T1])),
        (type_id(T2,Id2);print_error(['Unknown type ',T2])),
        matrix_element_at(Matrix,Id1,Id2,1),!,
        add_constraints_in_matrix(Rest, Matrix).

%% tests if a the type corresponding to indice Id1 is lesser than or equal to
%% the type corresponding to indice Id2
ordered_in_matrix(Matrix, Id1, Id2) :-
        matrix_element_at(Matrix, Id1, Id2, Element),
        Element == 1.

%% Now we build a data structure that will allow us to quickly determine
%% whenever there exists a type that have precise relation order w.r.t. other
%% types. This structures takes the form of a binary tree, each layer
%% corresponding to a precise type. Left part corresponds to types that are
%% greater (resp. lesser) than the type of the layer. Right part corresponds
%% to other types. This should accelerate the glb/lub lookup. This algorithm
%% is taken from the Gromit library included in the Wallace library written
%% by Francois Pottier (http://pauillac.inria.fr/~pottier), and was suggested
%% by Alexandre Frey.
build_lookup_trees(OrderMatrix, TreeUp, TreeDown) :-
        array_size(OrderMatrix,Size),
        verbose(debug_type_order,['get matrix size: ',Size]),
        build_lookup_trees(OrderMatrix, TreeUp, TreeDown, 1, Size).

build_lookup_trees(_,_,_,N,Size) :- N > Size ,!.
build_lookup_trees(OrderMatrix, TreeUp, TreeDown, Id, Size) :-
        build_lookup_trees(OrderMatrix, TreeUp, TreeDown, Id, 1, Size),
        Id1 is Id + 1,
        build_lookup_trees(OrderMatrix, TreeUp, TreeDown, Id1, Size).

build_lookup_trees(_,Id,Id,Id,N,Size) :-
        %% end of the tree: we put the ids at the leaves
        N > Size,!.
build_lookup_trees(OrderMatrix, t(UpLeft, UpRight), t(DownLeft, DownRight),
                   Id, Id2, Size) :-
        (
          ordered_in_matrix(OrderMatrix, Id2, Id)
        -> %% Id greater than Id2: choose left
          Up = UpLeft
        ;
          Up = UpRight
        ),
        (
          ordered_in_matrix(OrderMatrix, Id, Id2)
        -> %% Id lesser than Id2: choose left
          Down = DownLeft
        ;
          Down = DownRight
        ),
        Id21 is Id2 + 1,
        build_lookup_trees(OrderMatrix, Up, Down, Id, Id21, Size).

%% uses a previously built lookup tree to find a lub
lookup_lub(OrderMatrix, TreeDown, Ida, Idb, IdLub) :-
        lookup_lub(OrderMatrix, TreeDown, Ida, Idb, 1, IdLub).

lookup_lub(_, IdLub, _, _, _, IdLub) :- integer(IdLub),!.
lookup_lub(_, X,_,_,_,_) :- var(X),!,fail. % no lub
lookup_lub(OrderMatrix, t(Left,Right), Ida, Idb, IdLayer, IdLub) :-
        (
          ordered_in_matrix(OrderMatrix,Ida,IdLayer),
          ordered_in_matrix(OrderMatrix,Idb,IdLayer)
        ->
          NextNode = Left
        ;
          NextNode = Right
        ),
        NextLayer is IdLayer + 1,
        lookup_lub(OrderMatrix, NextNode, Ida, Idb, NextLayer, IdLub).

%% uses a previously built lookup tree to find a lub
lookup_glb(OrderMatrix, TreeUp, Ida, Idb, IdLub) :-
        lookup_glb(OrderMatrix, TreeUp, Ida, Idb, 1, IdLub).

lookup_glb(_,IdGlb, _, _, _, IdGlb) :- integer(IdGlb),!.
lookup_glb(_,X,_,_,_,_) :- var(X),!,fail. % no glb
lookup_glb(OrderMatrix, t(Left,Right), Ida, Idb, IdLayer, IdGlb) :-
        (
          ordered_in_matrix(OrderMatrix,IdLayer,Ida),
          ordered_in_matrix(OrderMatrix,IdLayer,Idb)
        ->
          NextNode = Left
        ;
          NextNode = Right
        ),
        NextLayer is IdLayer + 1,
        lookup_glb(OrderMatrix, NextNode, Ida, Idb, NextLayer, IdGlb).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Generate dynamic clauses
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% glb_clause(Tau1, Tau2, GLB, Clause)
%% Clause is a clause for predicate dyn_glb/3 telling
%% that
%% clp(Tau1, Tau2, GLB) is true if GLB is the Greatest Lower Bound of
%% of Tau1 and Tau2. Uses tclp__clp_types:tclp__vGLB to deals with the
%% arguments' GLBs
glb_clause(Tau1, Tau2, GLB, Clause) :-
        %% We forbid bottom here
        %% builtin_type(bot, Bot),
        %% GLB \= Bot,
                                % bot the now autorized, the check beeing done
                                % in the CHR solver
        arity(Tau1, N1), arity(Tau2, N2), arity(GLB, NGLB),
        functor(Type1, Tau1, N1),
        functor(Type2, Tau2, N2),
        functor(TypeGLB, GLB, NGLB),
        fold(Type1, Type2, TypeGLB,
             tclp__type_order:add_eq, tclp__type_order:add_vGLB, !, Body),
        Clause = (dyn_glb(Type1, Type2, TypeGLB) :- Body).

add_eq(A, B, Accu, ( Accu , A=B)).
add_vGLB(A,B,GLB, Accu, ( Accu , tclp__clp_types:tclp__vGLB(A,B,GLB))).

%% glb_clause_vars(Tau1, Tau2, GLB, Clause)
%% Clause is a clause for predicate dyn_glb_vars/3 telling
%% that if Tau1, Tau2 and GLB are type constructor with ord_sets of type
%% variables as their arguments, then the head of GLB is the glb of the
%% heads of Tau1 and Tau2, and its arguments are given by the union of
%% the corresponding arguments in Tau1 and Tau2.
glb_clause_vars(Tau1, Tau2, GLB, Clause) :-
        %% We forbid bottom here
        %% builtin_type(bot, Bot),
        %% GLB \= Bot,
                                % bot the now autorized, the check beeing done
                                % in the CHR solver        
        arity(Tau1, N1), arity(Tau2, N2), arity(GLB, NGLB),
        functor(Type1, Tau1, N1),
        functor(Type2, Tau2, N2),
        functor(TypeGLB, GLB, NGLB),
        fold(Type1, Type2, TypeGLB,
             tclp__type_order:add_eq, tclp__type_order:add_union, !, Body),
        Clause = (dyn_glb_vars(Type1, Type2, TypeGLB) :- Body).

add_union(A,B,GLB, Accu, ( Accu, ordsets:ord_union(A,B,GLB))).

%% glb_clause_id(Matrix, TreeUp, Id1, Id2, Clause,Clause2)
%% Clause is a clause for predicate dyn_glb/3 telling
%% that
%% clp(Tau1, Tau2, GLB) is true if GLB is the Greatest Lower Bound of
%% of Tau1 and Tau2 where Tau1 (resp. Tau2) is the type constructor with
%% assocaiated Id1 (resp. Id2). Uses tclp__clp_types:tclp__vGLB to deals
%% with the arguments' GLBs
%% Clause 2 is the corresponding clause for dyn_glb_vars/3
glb_clause_id(OrderMatrix, TreeUp, Id1, Id2, Clause, _Clause2) :-
        lookup_glb(OrderMatrix, TreeUp, Id1, Id2, IdGLB),
        id_type(Id1,Tau1), id_type(Id2, Tau2), id_type(IdGLB, GLB),
        glb_clause(Tau1, Tau2, GLB, Clause). % ,
%	glb_clause_vars(Tau1, Tau2, GLB, Clause2).

%% lub_clause(Tau1, Tau2, LUB, Clause)
%% Clause is a clause for predicate dyn_lub/3 telling
%% that
%% clp(Tau1, Tau2, LUB) is true if LUB is the Lowest Upper Bound of
%% of Tau1 and Tau2. Uses tclp__clp_types:tclp__vLUB to deals with the
%% arguments' LUBs
lub_clause(Tau1, Tau2, LUB, Clause) :-
        arity(Tau1, N1), arity(Tau2, N2), arity(LUB, NLUB),
        functor(Type1, Tau1, N1),
        functor(Type2, Tau2, N2),
        functor(TypeLUB, LUB, NLUB),
        fold(Type1, Type2, TypeLUB,
             tclp__type_order:add_eq, tclp__type_order:add_vLUB, !, Body),
        Clause = (dyn_lub(Type1, Type2, TypeLUB) :- Body).

add_vLUB(A,B,LUB, Accu, ( Accu , tclp__clp_types:tclp__vLUB(A,B,LUB))).
        
%% lub_clause_vars(Tau1, Tau2, LUB, Clause)
%% Clause is a clause for predicate dyn_lub_vars/3 telling
%% that if Tau1, Tau2 and LUB are type constructor with ord_sets of type
%% variables as their arguments, then the head of LUB is the lub of the
%% heads of Tau1 and Tau2, and its arguments are given by the union of
%% the corresponding arguments in Tau1 and Tau2.
lub_clause_vars(Tau1, Tau2, LUB, Clause) :-
        %% We forbid bottom here
        %% builtin_type(bot, Bot),
        %% LUB \= Bot,
                                % bot the now autorized, the check beeing done
                                % in the CHR solver
        arity(Tau1, N1), arity(Tau2, N2), arity(LUB, NLUB),
        functor(Type1, Tau1, N1),
        functor(Type2, Tau2, N2),
        functor(TypeLUB, LUB, NLUB),
        fold(Type1, Type2, TypeLUB,
             tclp__type_order:add_eq, tclp__type_order:add_union, !, Body),
        Clause = (dyn_lub_vars(Type1, Type2, TypeLUB) :- Body).


%% lub_clause_id(Matrix, TreeUp, Id1, Id2, Clause)
%% Clause is a clause for predicate dyn_lub/3 telling
%% that
%% clp(Tau1, Tau2, LUB) is true if LUB is the Greatest Lower Bound of
%% of Tau1 and Tau2 where Tau1 (resp. Tau2) is the type constructor with
%% assocaiated Id1 (resp. Id2). Uses tclp__clp_types:tclp__vLUB to deals
%% with the arguments' LUBs
%% Clause2 is the corresponding clause for predicate dyn_lub_vars/3
lub_clause_id(OrderMatrix, TreeDown, Id1, Id2, Clause, _Clause2) :-
        lookup_lub(OrderMatrix, TreeDown, Id1, Id2, IdLUB),
        id_type(Id1,Tau1), id_type(Id2, Tau2), id_type(IdLUB, LUB),
        lub_clause(Tau1, Tau2, LUB, Clause). % ,
%	lub_clause_vars(Tau1, Tau2, LUB, Clause2).


%% break_clause_no_check(Tau1,Tau2, Clause)
%% Clause is a clause for predicate dyn_break/2 telling
%% that
%% break(Tau1, Tau2) is true if Tau1 is lesser or equal to Tau2,
%% it uses tclp__clp_types:':<' to deal with arguments
%% The type constructors of Tau1 and Tau2 are supposed in the
%% right order here
break_clause_no_check(Tau1, Tau2, Clause) :-
        arity(Tau1, N1), arity(Tau2, N2), 
        functor(Type1, Tau1, N1),
        functor(Type2, Tau2, N2),
        fold(Type1, Type2, tclp__type_order:add_leq, !, Body),
        Clause = (dyn_break(Type1, Type2) :- Body).

add_leq(A,B,Accu, ( Accu , tclp__clp_types:':<'(A,B) , verbose(debug_clp_types,['break ',A,' :< ',B]))).

%% break_clause_id(OrderMatrix, Id1,Id2, Clause)
%% Clause is a clause for predicate dyn_break/3 telling
%% that break(Tau1, Tau2) is true if Tau1 is lesser or equal to Tau2,
%% where Tau1 (resp Tau2) is the type constructor associated with Id1
%% (resp Id2).
%% It uses tclp__clp_types:':<' to deal with arguments
break_clause_id(OrderMatrix, Id1, Id2, Clause) :-
        ordered_in_matrix(OrderMatrix, Id1, Id2),!,
        id_type(Id1,Tau1), id_type(Id2, Tau2),
        break_clause_no_check(Tau1, Tau2, Clause).

%% ordered_types_clause_no_check(Tau1, Tau2, Clause)
%% Clause is a clause for predicate dyn_ordered_types/2 telling
%% that ordered_types(Tau1,Tau2) is true if Tau1 < Tau2 is a direct
%% consequence of the current store of subtyping constraints
%% it uses tclp__clp_types:ordered_vars(_,_) to deal with arguments
%% The type constructors of Tau1 and Tau2 are supposed in the
%% right order here
ordered_types_clause_no_check(Tau1,Tau2,Clause) :-
        arity(Tau1, N1), arity(Tau2, N2), 
        functor(Type1, Tau1, N1),
        functor(Type2, Tau2, N2),
        fold(Type1, Type2, tclp__type_order:add_ordered_vars, !, Body),
        Clause = (dyn_ordered_types(Type1, Type2) :- Body).

add_ordered_vars(A,B,Accu, ( Accu, tclp__clp_types:ordered_vars(A,B) )).

%% ordered_types_clause_id(OrderMatrix, Id1, Id2, Clause)
%% Clause is a clause for predicate dyn_ordered_types/2 telling
%% that ordered_types(Tau1,Tau2) is true if Tau1 :< Tau2 is a direct
%% consequence of the current store of subtyping constraints
%% where Tau1 (resp Tau2) is the type constructor associated with Id1
%% (resp Id2).
ordered_types_clause_id(OrderMatrix, Id1, Id2, Clause) :-
        ordered_in_matrix(OrderMatrix, Id1, Id2),!,
        id_type(Id1,Tau1), id_type(Id2, Tau2),
        ordered_types_clause_no_check(Tau1,Tau2,Clause).

%% we use dynamic predicates to store LUBs, GLBs and Leq relations

:- dynamic dyn_glb/3, dyn_lub/3, dyn_break/2, dyn_ordered_types/2,
	dyn_glb_vars/3, dyn_lub_vars/3.

%% asserts the values for the two given constructors ids
assert_bounds_id(OrderMatrix, TreeUp, TreeDown, Id1, Id2) :-
        (
          glb_clause_id(OrderMatrix, TreeUp, Id1, Id2, GLBClause, _GLBClause2),
          asserta(GLBClause) % ,
	  % asserta(GLBClause2)
        ;
          true
        ),
        (
          lub_clause_id(OrderMatrix, TreeDown, Id1, Id2, LUBClause, _LUBClause2),
          asserta(LUBClause) % ,
	  % asserta(LUBClause2)
        ;
          true
        ),
        (
          break_clause_id(OrderMatrix, Id1, Id2, BreakClause),
          asserta(BreakClause)
        ;
          true
        ),
        (
          ordered_types_clause_id(OrderMatrix, Id1, Id2, CompareClause),
          asserta(CompareClause)
        ;
          true
        ),!.
 
assert_bounds_id(OrderMatrix, TreeUp, TreeDown, Id1, Id2, NbTypes) :-
        (
          Id1 == 0 -> true ; %% finished
          Id2 == 0 %% finished for Id1
        ->
          Id1b is Id1 - 1,
          assert_bounds_id(OrderMatrix, TreeUp, TreeDown,
                           Id1b, NbTypes, NbTypes)
        ;
          assert_bounds_id(OrderMatrix, TreeUp, TreeDown, Id1, Id2),
          Id2b is Id2 - 1,
          assert_bounds_id(OrderMatrix, TreeUp, TreeDown, Id1, Id2b, NbTypes)
        ).

%% (re)compute all bounds/break clauses for all types with method 2
build_order :-
        verbose(debug_type_order,'start build order'),
        init_order,
        verbose(debug_type_order,'init_order ok'),
        retractall(dyn_glb(_,_,_)),
        retractall(dyn_lub(_,_,_)),
        retractall(dyn_break(_,_)),
        retractall(dyn_ordered_types(_,_)),
        retractall(dyn_elim_arg_down(_,_,_)),
        retractall(dyn_elim_arg_up(_,_,_)),
        verbose(debug_type_order,'retractall order clauses ok'),
        associate_id_to_types(NbTypes),
        verbose(debug_type_order,['associate ids ok, ',NbTypes,' types']),
        make_order_matrix(NbTypes,OrderMatrix),
        verbose(debug_type_order,'order matrix ok'),
        build_lookup_trees(OrderMatrix, TreeUp, TreeDown),
        verbose(debug_type_order,'lookup trees ok'),
        assert_bounds_id(OrderMatrix, TreeUp, TreeDown,
                         NbTypes, NbTypes, NbTypes),
        verbose(debug_type_order,'assert order clauses ok'),
        assert_elim_clauses(OrderMatrix),
        verbose(debug_type_order,'assert elim clauses ok'),
        build_extrema_enumeration,
        verbose(debug_type_order,'build_order done').

%% access : non dynamic predicates that allow to access to the value of the 
%% dynamic ones. theses predicates are the one that are exported

%% tclp__glb(Type1, Type2, GLB) poses the constraint that GLB the is
%% GLB of Type1 and Type2
tclp__glb(Type1, Type2, GLB) :-
        dyn_glb(Type1, Type2, GLB),!.
%% handling type variable constants
tclp__glb(Type1, Type2, GLB) :-
        atom(Type1), Type1==Type2, !, GLB=Type2.
tclp__glb(Type1, Top, Type1) :-
        atom(Type1), atom(Top), builtin_type(top,Top2), Top2==Top,!.
tclp__glb(Top, Type2, Type2) :-
        atom(Type2), atom(Top), builtin_type(top,Top2), Top2==Top,!.
%% verbose failure
tclp__glb(Type1, Type2, _) :-
        verbose(debug_type_order,['glb failed for ',Type1,' , ',Type2]),
        fail.

%% tclp__glb_vars(List,TGLB)
%% List ids a list of type constructors with ordsets of variables as arguments
%% TGLB head constructor is the GLB of List with its arguments being the
%% union of the corresponding arguments in the elements of list
tclp__glb_vars([X|Rest],TGLB) :- tclp__glb_vars_aux(Rest,X,TGLB).

tclp__glb_vars_aux([],TGLB,TGLB).
tclp__glb_vars_aux([X|Rest],Accu,TGLB) :-
	tclp__glb_vars_aux2(X,Accu,Accu2),
	tclp__glb_vars_aux(Rest,Accu2,TGLB).

tclp__glb_vars_aux2(X,Y,GLB) :-
	dyn_glb_vars(X,Y,GLB),!.
tclp__glb_vars_aux2(X,Y,GLB) :-
	atom(X), X==Y, !, GLB=X.
tclp__glb_vars_aux2(Type1, Top, Type1) :-
        atom(Type1), atom(Top), builtin_type(top,Top2), Top2==Top,!.
tclp__glb_vars_aux2(Top, Type2, Type2) :-
        atom(Type2), atom(Top), builtin_type(top,Top2), Top2==Top,!.
tclp__glb_vars_aux2(Type1, Type2, _) :-
        verbose(debug_type_order,['glb failed for ',Type1,' , ',Type2]),
        fail.

%% tclp__lub(Type1, Type2, LUB) poses the constraint that LUB the is
%% LUB of Type1 and Type2
tclp__lub(Type1, Type2, LUB) :-
        dyn_lub(Type1, Type2, LUB),!.
%% handling type variable constants
tclp__lub(Type1, Type2, LUB) :-
        atom(Type1), Type1==Type2, !, LUB=Type2.
tclp__lub(Type1, Bot, Type1) :-
        atom(Type1), atom(Bot), builtin_type(bot,Bot2), Bot2==Bot,!.
tclp__lub(Bot, Type2, Type2) :-
        atom(Type2), atom(Bot), builtin_type(bot,Bot2), Bot2==Bot,!.
tclp__lub(Top, Type, Top) :-
        atom(Type), atom(Top), builtin_type(top,Top2), Top2==Top,!.
tclp__lub(Type,Top,Top) :-
        atom(Type), atom(Top), builtin_type(top,Top2), Top2==Top,!.
%% verbose failure
tclp__lub(Type1, Type2, _) :-
        verbose(debug_type_order,['lub failed for ',Type1,' , ',Type2]),
        fail.

%% tclp__break(Type1, Type2) adds the constraint that Type1 is smaller or 
%% equal to Type2
tclp__break(Type1, Type2) :-
        dyn_break(Type1, Type2),!.
%% handling type variable constants
tclp__break(Type1, Type2) :-
        atom(Type1), atom(Type2),
        (
          Type1==Type2,!
        ;
          builtin_type(bot, Bot), Type1==Bot,!
        ;
          builtin_type(top, Top), Type2==Top
        ),!.
%% verbose failure
tclp__break(Type1, Type2) :-
        verbose(debug_type_order,['break failed for ',Type1,' , ',Type2]),
        fail.

%% tclp__ordered_types(Type1, Type2) check if Type1 :< Type2 is a direct
%% consequence of the current store of subtyping constraints
tclp__ordered_types(Type1, Type2) :-
        dyn_ordered_types(Type1, Type2),!.
%% handling type variable constants
tclp__ordered_types(Type1, Type2) :-
        atom(Type1), atom(Type2),
        (
          Type1==Type2,!
        ;
          builtin_type(bot, Bot), Type1==Bot,!
        ;
          builtin_type(top, Top), Type2==Top
        ),!.


%% add_type(Type)
%% Adds a type to the base
add_type(Type) :-
        functor(Type, Cons, NArgs),
        add_type_memo(Cons,NArgs).

%% add_type(Type)
%% Adds a type to the base
add_type(Cons, NArgs) :-
        add_type_memo(Cons,NArgs).
add_type_chr(Cons, NArgs) :-
        type(Cons),
        (arity(Cons,NArgs) -> true ;
            arity(Cons,N),
            print_error(['Type constructor "',Cons,
                         '" was declared twice with two different number of arguments: ',
                         N,' and ', NArgs]) ).

%% add_order(Type1, Type2)
%% adds in the relation that Type1 is smaller or equal to Type2
add_order(Type1, Type2) :-
        add_order_memo(Type1, Type2).
add_order_chr(Type1, Type2) :-
        functor(Type1, Cons1, _),
        functor(Type2, Cons2, _),
        lt(Cons1, Cons2),
        unifyToArity(Type1),
        unifyToArity(Type2).

%% type_exists(type)
%% checks if a type is well defined (all its components must have been
%% declared)
type_exists(X) :- var(X),!.
type_exists(Type) :-
        functor(Type,Head,NArgs),
        memo_type(Head,NArgs),
        type_exists(Type,NArgs).
type_exists(_,0) :- !.
type_exists(Type,N) :-
        arg(N,Type,Arg),
        type_exists(Arg),
        N1 is N-1,
        type_exists(Type,N1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Memoization of types and order declarations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic memo_type/2, memo_order/2.

%% Adds a type to the base whenever it is needed
add_type_memo(T,0) :-
        builtin_type(term,Term), T==Term, option(disableterm),!.
                                % don't add term if the disableterm option
                                % is passed
add_type_memo(Type,Arity) :-
        memo_type(Type,Arity),!.
add_type_memo(Type,Arity) :-
        assert(memo_type(Type,Arity)).

%% adds an order relation
add_order_memo(Type1,Type2) :-
        assert(memo_order(Type1,Type2)).

%% adds alls the given types with their arity to the chr handler using
%% add_type_chr/2.
add_all_types_chr([]).
add_all_types_chr([Type-NArgs|Rest]) :-
        add_type_chr(Type,NArgs),
        add_all_types_chr(Rest).
add_all_types_chr :-
        findall(T-A,memo_type(T,A),List),
        verbose(debug_type_order,['find types ok']),
        add_all_types_chr(List).

%% adds alls the given order declarations to the chr handler using
%% add_order_chr/2 .
add_all_order_decs([]).
add_all_order_decs([Type1-Type2|Rest]) :-
        add_order_chr(Type1,Type2),
        add_all_order_decs(Rest).
add_all_order_decs :-
        findall(T1-T2,memo_order(T1,T2),List),
        add_all_order_decs(List).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Complete the order with maximal and minimal type that have no arity
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% true if some given type is a maximal type
is_maximal(Type) :-
        chr:findall_constraints(lt(Type,_),List),
        builtin_type(top,Top), 
        \+ (member(lt(Type,T)#_,List), T\=Top).

%% true if the given type constructor is a minimal type
is_minimal(Type) :-
        chr:findall_constraints(lt(_,Type),List),
        builtin_type(bot,Bot),
        \+ (member(lt(T,Type)#_,List), T\=Bot).

%% find the current maximal types of the order
find_maximals(Maximals) :-
        findall(T,
                ( memo_type(T,_),is_maximal(T) ),
                List),
        builtin_type(top,Top),
        delete(List,Top,Maximals).

find_minimals(Minimals) :-
        findall(T,
                ( memo_type(T,_), is_minimal(T)),
                List),
        builtin_type(bot,Bot),
        delete(List,Bot,Minimals).

%% complete the maximals with empty types when needed
complete_maximals([]).
complete_maximals([T|Rest]) :-
        arity(T,N),
        N > 0,!,
        atom_concat('$empty_up_',T,NT),
        add_type(NT), functor(Type,T,N),
        add_order(Type,NT),
        complete_maximals(Rest).
complete_maximals([_|Rest]) :-
        complete_maximals(Rest).

%% complete the minimals with empty types when needed
complete_minimals([]).
complete_minimals([T|Rest]) :-
        arity(T,N), N > 0,!,
        atom_concat('$empty_down_',T,NT),
        add_type(NT), functor(Type,T,N),
        add_order(NT,Type),
        complete_minimals(Rest).
complete_minimals([_|Rest]) :-
        complete_minimals(Rest).

%% make the current order completion
complete_order :-
        find_maximals(Maxi),
        complete_maximals(Maxi). %,
%        find_minimals(Mini),
%        complete_minimals(Mini).

%% predicates that enumerate maximas and minimas
:- dynamic tclp__enumerate_tops/1, tclp__enumerate_bottoms/1.

build_extrema_enumeration :-
        find_maximals(Maxi),
        retractall(tclp__enumerate_tops(_)),
        findall(X,
                ((lists:member(X,Maxi);dyn_type_variable_constant(X)),
                    assertz(tclp__enumerate_tops(X))),
                _),
        find_minimals(Mini),
        retractall(tclp__enumerate_bottoms(_)),
        findall(Y,
                ((lists:member(Y,Mini);dyn_type_variable_constant(Y)),
                    assertz(tclp__enumerate_bottoms(Y))),
                _).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Argument elimination
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% true if the given type constructor C whose id is Id has X among the
%% labels of its arguments
has_label(Id,X) :-
        id_type(Id,C),
        memo_type(C,N), functor(Type,C,N),
        unifyToArity(Type), Type =.. [C|Labels],
        true_member(X,Labels),!.

%% find for each type constructor and each argument the type constructor
%% which forgets this argument downward

find_elim_down_id(Matrix, OrigId, ArgNo, NewId) :-
        builtin_type(bot,Bot),
        type_id(Bot,BId), id_type(OrigId, OrigCons),
        matrix_size(Matrix, Size, Size),
        arity(X,OrigCons,ArgNo),
        find_elim_down_id2(Matrix, OrigId, X, 1, Size, BId, NewId).
find_elim_down_id2(_Matrix, _OrigId, _X,
                   Current, Size, CurrentResult, NewId) :-
        Current > Size, !, CurrentResult=NewId.
find_elim_down_id2(Matrix, OrigId, X, Current, Size, CurrentResult, NewId) :-
        ordered_in_matrix(Matrix, Current, OrigId),
                                % Current is under the original
        ordered_in_matrix(Matrix, CurrentResult, Current),
                                % and is also a better candidate than the
                                % current one (since candidates form a
                                % quasi-lattice, we don't need to keep a set
                                % of possible maximal candiates)
        \+ has_label(Current,X), % this type constructor is a candidate
        !,                      % we found a better candidate!
        Next is Current +1,
        find_elim_down_id2(Matrix, OrigId, X, Next, Size, Current, NewId).
                                % finally we replace CurrentResult by Current
                                % in the recursive call
find_elim_down_id2(Matrix, OrigId, X, Current, Size, CurrentResult, NewId) :-
        Next is Current + 1,
        find_elim_down_id2(Matrix, OrigId, X,
                           Next, Size, CurrentResult, NewId).

%% find for each type constructor and each argument the type constructor
%% which forgets this argument upward

find_elim_up_id(Matrix, OrigId, ArgNo, NewId) :-
        builtin_type(bot,Bot),
        type_id(Bot,BId), id_type(OrigId, OrigCons),
        matrix_size(Matrix, Size, Size),
        arity(X,OrigCons,ArgNo),
        find_elim_up_id2(Matrix, OrigId, X, 1, Size, BId, NewId).
find_elim_up_id2(_Matrix, _OrigId, _X,
                   Current, Size, CurrentResult, NewId) :-
        Current > Size, !, CurrentResult=NewId.
find_elim_up_id2(Matrix, OrigId, X, Current, Size, CurrentResult, NewId) :-
        ordered_in_matrix(Matrix, OrigId, Current),
                                % Current is over the original
        ordered_in_matrix(Matrix, Current, CurrentResult),
                                % and is also a better candidate than the
                                % current one (since candidates form a
                                % quasi-lattice, we don't need to keep a set
                                % of possible maximal candiates)
        \+ has_label(Current,X),
                                % this type constructor is a candidate
        !,                      % we found a better candidate!
        Next is Current +1,
        find_elim_up_id2(Matrix, OrigId, X, Next, Size, Current, NewId).
                                % finally we replace CurrentResult by Current
                                % in the recursive call
find_elim_up_id2(Matrix, OrigId, X, Current, Size, CurrentResult, NewId) :-
        Next is Current + 1,
        find_elim_up_id2(Matrix, OrigId, X, Next, Size, CurrentResult, NewId).

%% we use dynamic predicates to proceed to argument elimination

:- dynamic dyn_elim_arg_down/3, dyn_elim_arg_up/3.

%% makes a goal that equalize the corresponding arguments of two types
make_eq_type_goal(Type1, Type2, Goal) :-
        fold(Type1, Type2, tclp__type_order:add_eq, !, Goal).

%% asserts the needed clauses for argument elimination
assert_elim_clauses(Matrix) :-
        matrix_size(Matrix, Size, Size),
        assert_elim_clauses2(Matrix, 1, Size).
assert_elim_clauses2(_Matrix, I, Size) :- I > Size,!.
assert_elim_clauses2(Matrix, I, Size) :-
        id_type(I,C), arity(C,N),
        assert_elim_clauses3(Matrix, I, 1, N),
        Next is I+1,
        assert_elim_clauses2(Matrix, Next, Size).
assert_elim_clauses3(_Matrix, _Id, ArgNo, NbArgs) :- ArgNo > NbArgs,!.
assert_elim_clauses3(Matrix, Id, ArgNo, NbArgs) :-
        find_elim_down_id(Matrix, Id, ArgNo, DownId),
        find_elim_up_id(Matrix, Id, ArgNo, UpId),
        id_type(Id, Tau), id_type(DownId, DownTau), id_type(UpId, UpTau),
        arity(DownTau,DownNbArgs), arity(UpTau, UpNbArgs),
        functor(Type,Tau,NbArgs), functor(DownType, DownTau, DownNbArgs),
        functor(UpType, UpTau, UpNbArgs),
        builtin_type(bot,Bot), builtin_type(top,Top),
        ( DownType == Bot ->
            DownGoal = (!,fail) ;
            make_eq_type_goal(Type, DownType, DownGoal) ),
        ( UpType == Top ->
            UpGoal = (!,fail) ;
            make_eq_type_goal(Type, UpType, UpGoal) ),
        assert((dyn_elim_arg_down(Type, ArgNo, DownType) :- DownGoal)),
        assert((dyn_elim_arg_up(Type, ArgNo, UpType) :- UpGoal)),!.

tclp__elim_arg_up(Type,No,NewType) :- dyn_elim_arg_up(Type,No,NewType).
tclp__elim_arg_down(Type,No,NewType) :- dyn_elim_arg_down(Type,No,NewType).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% dealing with type variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic dyn_type_variable_constant/1.

add_type_variable_constant(Name) :-
        dyn_type_variable_constant(Name),!.
add_type_variable_constant(Name) :-
        assert(dyn_type_variable_constant(Name)).
is_type_variable_constant(Name) :-
	dyn_type_variable_constant(Name),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Reinitialisation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

reinit :-
        retractall(type_id(_,_)),
        retractall(id_type(_,_)),
        retractall(dyn_glb(_,_,_)),
        retractall(dyn_lub(_,_,_)),
        retractall(dyn_break(_,_)),
        retractall(memo_type(_,_)),
        retractall(memo_order(_,_)),
        remove_all_constraints(lt(_,_)),
        remove_all_constraints(type(_)).
